from vllm.v1.worker.cpu_model_runner import CPUModelRunner


class TorchSpyreModelRunner(CPUModelRunner):
    """Torch Spyre specific model runner class inheriting from the CPUModelRunner."""

    raise NotImplementedError
